from typing import Any, Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from zuper_typing import dataclass


import numpy as np

from zuper_typing.literal import make_Literal  # XXX
from .basics import InteractionProtocol
from .protocol_agent import EpisodeStart

__all__ = [
    "RobotState",
    "RobotName",
    "RobotObservations",
    "RobotInterfaceDescription",
    "RobotConfiguration",
    "RobotPerformance",
    "ScenarioRobotSpec",
    "SetRobotCommands",
    "GetRobotObservations",
    "GetRobotState",
    "Metric",
    "PerformanceMetrics",
    "protocol_simulator",
    "protocol_scenario_maker",
    "SetMap",
    "SpawnRobot",
    "Step",
    "SimulationState",
    "Scenario",
    "ScenarioDuckieSpec",
    "StateDump",
    "JPGImage",
    "DumpState",
    "MOTION_PARKED",
    "MOTION_MOVING",
]

RobotName = str


@dataclass
class JPGImage:
    """
        An image in JPG format.

        jpg_data: Bytes of a JPG file
    """

    jpg_data: bytes


@dataclass
class SetMap:
    map_data: Any


@dataclass
class RobotConfiguration:
    pose: np.ndarray
    velocity: np.ndarray


@dataclass
class SetRobotCommands:
    robot_name: RobotName
    t_effective: float
    commands: Any


@dataclass
class GetRobotObservations:
    robot_name: RobotName
    t_effective: float


@dataclass
class GetRobotState:
    robot_name: RobotName
    t_effective: float


@dataclass
class RobotObservations:
    robot_name: RobotName
    t_effective: float
    observations: Any


@dataclass
class RobotState:
    robot_name: RobotName
    t_effective: float
    state: Any


@dataclass
class SimulationState:
    """
        Returns the simulation state.

        done: Whether the simulation should be terminated.
        done_why: Human-readable short message.
        done_code: Short string to use as code for statistics.
    """

    done: bool
    done_why: Optional[str]
    done_code: Optional[str]


@dataclass
class Metric:
    higher_is_better: bool
    cumulative_value: float
    description: str

    def __post_init__(self):
        if isinstance(self.cumulative_value, int):
            self.cumulative_value = float(self.cumulative_value)


@dataclass
class PerformanceMetrics:
    """
        Performance metrics for an agent.

        By convention there will be one called "reward" for RL tasks.

        Note that the values are *cumulative* to make it possible to have
        a sampling-invariant behavior.
    """

    metrics: Dict[str, Metric]


@dataclass
class RobotPerformance:
    robot_name: RobotName
    t_effective: float
    performance: PerformanceMetrics


@dataclass
class RobotInterfaceDescription:
    robot_name: RobotName
    observations: type
    commands: type


@dataclass
class Step:
    until: float


@dataclass
class DumpState:
    pass


@dataclass
class StateDump:
    """ Opaque object that contains the simulator's state, whichever it is. """

    state: object


MOTION_PARKED = "parked"
MOTION_MOVING = "moving"
NPMotion = make_Literal(MOTION_PARKED, MOTION_MOVING)


@dataclass
class ScenarioRobotSpec:
    description: str
    playable: bool
    configuration: RobotConfiguration
    color: str

    # if not playable
    motion: Optional[NPMotion]


@dataclass
class ScenarioDuckieSpec:
    color: str  # css color
    pose: np.ndarray  # SE2


@dataclass
class SpawnRobot:
    playable: bool
    robot_name: RobotName
    configuration: RobotConfiguration
    motion: Optional[NPMotion]


@dataclass
class Scenario:
    # Specification of the environments
    scenario_name: str
    environment: Any
    robots: Dict[str, ScenarioRobotSpec]
    duckies: Dict[str, ScenarioDuckieSpec]


description = """\

A "scenario maker" is an object that is able to create robot simulations scenarios.

Each scenario contains:

- The definition of a map and other environmental conditions.
- The pose of the "playing" robot. 


A scenario maker can be random. In that case it must give the same sequence
for the same values of the seed. 

"""

protocol_scenario_maker = InteractionProtocol(
    description=description,
    language="""\
        in:seed? ;
        (
            in:next_scenario ; 
            (out:finished | out:scenario)
        )*
""",
    inputs={
        # Seed random number generator
        "seed": int,
        "next_scenario": type(None),
    },
    outputs={"finished": type(None), "scenario": Scenario},
)
description = """\

Interface to be implemented by a simulator.

Logical API:

`simulator.clear(reset_info)`

Resets the simulation data. Need to re-transmit map and robot poses.

`simulator.set_map(map)`

Sets the map to use.

`simulator.spawn_robot(name, configuration)`

Adds a robot to the simulation of the given name.

`simulator.episode_start`

`simulator.step(until: timestamp)`

Steps the simulation until the given timestamp.


`simulator.set_robot_commands(t: timestamp, commands)`

Set the robot commands.


`simulator.get_robot_observations(name, t: timestamp)`

Asks for the dump of a robot observation.


`simulator.dump_robot_state(name)`

Asks for the dump of a robot state.


`seed(int)`

Sets seed for random process.


    """

protocol_simulator = InteractionProtocol(
    description=description,
    language="""\
        
        in:seed? ;
            (
            in:clear ; 
            in:set_map ;
            (in:spawn_robot)*;
            
            (in:get_robot_interface_description; out:robot_interface_description)*;
            
            in:episode_start;
                
                (
                    in:step |
                    in:set_robot_commands |
                    (in:get_robot_observations ;  out:robot_observations) |
                    (in:get_robot_performance ;  out:robot_performance) |
                    (in:get_robot_state ;  out:robot_state) |
                    (in:get_sim_state ;  out:sim_state) |
                    (in:dump_state   ;  out:state_dump) |
                    (in:get_ui_image ; out:ui_image) 
                )* 
        )*
""",
    inputs={
        # Seed random number generator
        "seed": int,
        "clear": type(None),
        "set_map": SetMap,
        "spawn_robot": SpawnRobot,
        "get_robot_interface_description": RobotName,
        "get_robot_performance": RobotName,
        "get_sim_state": type(None),
        "episode_start": EpisodeStart,
        # Step physics
        "step": Step,
        "set_robot_commands": SetRobotCommands,
        "get_robot_observations": GetRobotObservations,
        "get_robot_state": GetRobotState,
        "get_ui_image": type(None),
        # Dump state information
        "dump_state": DumpState,
    },
    outputs={
        "robot_observations": RobotObservations,
        "robot_state": RobotState,
        "robot_performance": RobotPerformance,
        "robot_interface_description": RobotInterfaceDescription,
        "sim_state": SimulationState,
        "state_dump": StateDump,
        "ui_image": JPGImage,
    },
)

# XXX this is not implemented yet
protocol_scorer = InteractionProtocol(
    description="""Protocol for scorer""",
    language="""\
            in:set_map ;
            (
                in:episode_start;
                (
                    in:set_sim_state |
                    in:set_robot_state |
                    (in:get_robot_performance; out:robot_performance)
                )*
            )*
""",
    inputs={
        "set_map": SetMap,
        "episode_start": EpisodeStart,
        "set_robot_state": RobotState,
        "set_sim_state": SimulationState,
        "get_robot_performance": RobotName,
    },
    outputs={"robot_performance": RobotPerformance},
)



#!/usr/bin/env python3
from aido_schemas import protocol_image_filter, JPGImage, EpisodeStart, wrap_direct, Context


class DummyImageFilter:
    def init(self, context: Context):
        context.info("init()")

    def on_received_episode_start(self, context: Context, data: EpisodeStart):
        context.write("episode_start", data=data)

    def on_received_image(self, context: Context, data: JPGImage):
        transformed = data
        context.write("image", transformed)

    def finish(self, context: Context):
        context.info("finish()")


def main():
    node = DummyImageFilter()
    protocol = protocol_image_filter
    wrap_direct(node=node, protocol=protocol)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import time
from contextlib import contextmanager

from dataclasses import dataclass, field
from typing import Dict

__all__ = ["TimeTracker"]


@dataclass
class TimeTracker:
    step: int
    total: float = 0
    phases: Dict[str, float] = field(default_factory=dict)

    @contextmanager
    def measure(self, phase_name: str):
        t0 = time.time()
        yield
        t1 = time.time()
        delta = t1 - t0
        self.phases[phase_name] = delta
        self.total += delta



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from zuper_typing import dataclass

from .basics import InteractionProtocol

__all__ = ["EpisodeStart", "protocol_agent", "GetCommands"]


@dataclass
class EpisodeStart:
    """ Marker for the start of an episode. """

    episode_name: str


@dataclass
class GetCommands:
    at_time: float


protocol_agent = InteractionProtocol(
    description="""

Generic protocol for an agent that receives "observations" and responds 
with "commands".

"episode_start" marks the beginning of an episode.  

    """.strip(),
    inputs={"observations": Any, "seed": int, "get_commands": GetCommands, "episode_start": EpisodeStart,},
    outputs={"commands": Any},
    language="""
            in:seed? ;
            (   in:episode_start ; 
                (in:observations | 
                    (in:get_commands ; out:commands)
                 )* 
            )*
        """,
)



#!/usr/bin/env python3
from dataclasses import dataclass, field
from typing import Tuple

# noinspection PyUnresolvedReferences
import cv2
import numpy as np


from aido_schemas import protocol_image_source, JPGImage, EpisodeStart, wrap_direct, Context


# noinspection PyUnresolvedReferences
from zuper_nodes.structures import timestamp_from_seconds, TimeSpec, TimingInfo


@dataclass
class DummyImageSourceConfig:
    """
        Configuration for the node.

        :param shape: Image shape in Numpy conventions (height, width).
        :param images_per_episode: Number of images for each episode.
        :param num_episodes: Number of episodes in total.
    """

    shape: Tuple[int, int] = (480, 640)
    images_per_episode: int = 120
    num_episodes: int = 10


@dataclass
class DummyImageSourceState:
    """
        State for the node.
    """

    episode: int = -1
    nimages: int = -1

    episode_name: str = None


@dataclass
class DummyImageSource:
    """ A simple example of an image source """

    config: DummyImageSourceConfig = field(default_factory=DummyImageSourceConfig)
    state: DummyImageSourceState = field(default_factory=DummyImageSourceState)

    def init(self):
        self.state.episode = -1
        self.state.nimages = -1

    def on_updated_config(self, context: Context, key: str, value):
        context.info(f"Config was updated: {key} = {value!r}")

    def on_received_next_episode(self, context: Context):
        self._start_episode(context)

    def on_received_next_image(self, context: Context):
        if self.state.nimages >= self.config.images_per_episode:
            context.write("no_more_images", None)
            return

        H, W = self.config.shape
        values = (128 + np.random.randn(H, W, 3) * 60).astype("uint8")
        jpg_data = bgr2jpg(values)
        image = JPGImage(jpg_data)
        delta = 0.15
        t = self.state.nimages * delta
        time = timestamp_from_seconds(t)
        ts = TimeSpec(time=time, frame=self.state.episode_name, clock=context.get_hostname())
        acquired = {"image": ts}
        timing = TimingInfo(acquired=acquired)
        context.write(topic="image", data=image, timing=timing)

    def finish(self):
        pass

    def _start_episode(self, context: Context):
        if self.state.episode >= self.config.num_episodes:
            context.write("no_more_episodes", None)
            return

        self.state.episode += 1
        self.state.nimages = 0
        self.state.episode_name = f"episode{self.state.episode}"

        context.write("episode_start", EpisodeStart(self.state.episode_name))


def bgr2jpg(image_cv) -> bytes:

    compress = cv2.imencode(".jpg", image_cv)[1]
    jpg_data = np.array(compress).tobytes()
    return jpg_data


def main():
    node = DummyImageSource()
    protocol = protocol_image_source
    wrap_direct(node=node, protocol=protocol)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import numpy as np

from aido_schemas import (
    JPGImage,
    Duckiebot1Observations,
    Duckiebot1Commands,
    wrap_direct,
    Context,
    SetMap,
    SpawnRobot,
    RobotName,
    StateDump,
    Step,
    RobotInterfaceDescription,
    RobotState,
    protocol_simulator_duckiebot1,
    RobotPerformance,
    Metric,
    PerformanceMetrics,
    DB18SetRobotCommands,
    DB18RobotObservations,
)


class DummySimulator:
    """ A dummy simulator implementation. """

    current_time: float
    robot_name: str

    def init(self, context: Context):
        context.info("init()")

    def on_received_seed(self, context: Context, data: int):
        context.info(f"seed({data})")

    def on_received_clear(self, context: Context):
        context.info(f"clear()")

    def on_received_set_map(self, context: Context, data: SetMap):
        context.info(f"set_map({data})")
        # TODO: load map

    def on_received_spawn_robot(self, data: SpawnRobot):
        self.robot_name = data.robot_name
        # TODO: set pose of robot

    def on_received_get_robot_interface_description(self, context: Context, data: RobotName):
        rid = RobotInterfaceDescription(
            robot_name=data, observations=Duckiebot1Observations, commands=Duckiebot1Commands
        )
        context.write("robot_interface_description", rid)

    def on_received_get_robot_performance(self, context: Context, data: RobotName):
        context.info(f"get_robot_interface_description()")
        metrics = {}
        metrics["reward"] = Metric(
            higher_is_better=True,
            cumulative_value=self.current_time,
            description="Dummy reward equal to survival time.",
        )
        pm = PerformanceMetrics(metrics)
        rid = RobotPerformance(robot_name=data, t_effective=self.current_time, performance=pm)
        context.write("robot_performance", rid)

    def on_received_episode_start(self, context: Context):
        context.info(f"episode_start()")
        self.current_time = 0

    def on_received_step(self, context: Context, data: Step):
        context.info(f"step({data})")
        self.current_time = data.until

    def on_received_set_robot_commands(self, context: Context, data: DB18SetRobotCommands):
        context.info(f"set_robot_commands({data})")

    def on_received_get_robot_observations(self, context, data: RobotName):
        context.log(f"get_robot_observation({data!r})")
        camera = get_random_image(shape=(200, 300))
        obs = Duckiebot1Observations(camera)
        ro = DB18RobotObservations(
            robot_name=self.robot_name, t_effective=self.current_time, observations=obs
        )
        context.write("robot_observations", ro, with_schema=True)

    def on_received_get_robot_state(self, context: Context, data: RobotName):
        context.info(f"get_robot_state({data!r})")
        rs = RobotState(robot_name=data, t_effective=self.current_time, state=None)
        context.write("robot_state", rs)

    def on_received_dump_state(self, context: Context):
        context.info(f"dump_state()")
        context.write("dump_state", StateDump(None))


def get_random_image(shape):
    H, W = shape
    values = (128 + np.random.randn(H, W, 3) * 60).astype("uint8")
    jpg_data = bgr2jpg(values)
    image = JPGImage(jpg_data)
    return image


import cv2

# noinspection PyUnresolvedReferences
def bgr2jpg(image_cv) -> bytes:
    compress = cv2.imencode(".jpg", image_cv)[1]
    jpg_data = np.array(compress).tostring()
    return jpg_data


def main():
    node = DummySimulator()
    protocol = protocol_simulator_duckiebot1
    wrap_direct(node=node, protocol=protocol)


if __name__ == "__main__":
    main()



from setuptools import setup


def get_version(filename: str):
    import ast

    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename="src/aido_schemas/__init__.py")

line = "daffy"
install_requires = [
    "compmake-z6",
    "pyparsing",
    "PyContracts3",
    "networkx>=2,<3",
    "termcolor",
    "pydot",
    "zuper-ipce-z6>=6.0.34",
    "zuper-nodes-z6>=6.0.33",
]

setup(
    name=f"aido-protocols-{line}",
    version=version,
    keywords="",
    package_dir={"": "src"},
    packages=["aido_schemas", "aido_schemas_tests",],
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "aido-log-draw=aido_schemas.utils_drawing:aido_log_draw_main",
            "aido-log-video=aido_schemas.utils_video:aido_log_video_main",
            "aido-log-video-ui-image=aido_schemas.utils_video:aido_log_video_ui_image_main",
        ],
    },
)



__version__ = "6.0.20"

from zuper_commons.logs import ZLogger

logger = ZLogger(__name__)
logger.info(f"aido-protocols {__version__}")

from .protocols import *
from .protocol_agent import *
from .protocol_simulator import *
from .schemas import *
from .basics import *



#!/usr/bin/env python3

import numpy as np

from aido_schemas import (
    EpisodeStart,
    protocol_agent_DB20,
    PWMCommands,
    Duckiebot1Commands,
    LEDSCommands,
    RGB,
    wrap_direct,
    Context,
)


class RandomAgent:
    def init(self, context: Context):
        self.n = 0
        context.info("init()")

    def on_received_seed(self, data: int):
        np.random.seed(data)

    def on_received_episode_start(self, context: Context, data: EpisodeStart):
        context.info(f'Starting episode "{data.episode_name}".')

    def on_received_observations(self, context: Context):
        pass

    def on_received_get_commands(self, context: Context):
        if self.n == 0:
            pwm_left = 0.0
            pwm_right = 0.0
        else:
            pwm_left = np.random.uniform(0.5, 1.0)
            pwm_right = np.random.uniform(0.5, 1.0)
        self.n += 1

        # pwm_left = 1.0
        # pwm_right = 1.0
        grey = RGB(0.0, 0.0, 0.0)
        led_commands = LEDSCommands(grey, grey, grey, grey, grey)
        pwm_commands = PWMCommands(motor_left=pwm_left, motor_right=pwm_right)
        commands = Duckiebot1Commands(pwm_commands, led_commands)
        context.write("commands", commands)

    def finish(self, context: Context):
        context.info("finish()")


def main():
    node = RandomAgent()
    protocol = protocol_agent_DB20
    wrap_direct(node=node, protocol=protocol)


if __name__ == "__main__":
    main()



from typing import Dict, TYPE_CHECKING

import numpy as np
from zuper_nodes import particularize_no_check

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from zuper_typing import dataclass

from .protocol_agent import protocol_agent
from .protocol_simulator import (
    JPGImage,
    protocol_simulator,
    RobotName,
    RobotObservations,
    RobotState,
    SetRobotCommands,
    StateDump,
)

__all__ = [
    "PWMCommands",
    "Duckiebot1Observations",
    "Duckiebot1ObservationsPlusState",
    "Duckiebot1Commands",
    "LEDSCommands",
    "RGB",
    "DB18SetRobotCommands",
    "DB18RobotObservations",
    "protocol_agent_duckiebot1",
    "protocol_simulator_duckiebot1",
    "protocol_agent_duckiebot1_fullstate",
    "DTSimRobotInfo",
    "DTSimRobotState",
    "DTSimState",
    "DTSimStateDump",
    "DB20Observations",
    "DB20Commands",
    "DB20Odometry",
    "DB20ObservationsPlusState",
    "DB20RobotObservations",
    "DB20SetRobotCommands",
    "protocol_agent_DB20",
    "protocol_agent_DB20_fullstate",
    "protocol_simulator_DB20",
]


@dataclass
class PWMCommands:
    """
        PWM commands are floats between -1 and 1.
    """

    motor_left: float
    motor_right: float

    def __post_init__(self):
        self.motor_left = float(self.motor_left)
        self.motor_right = float(self.motor_right)
        m = max(abs(self.motor_left), abs(self.motor_right))
        if m > 1:
            msg = f"Expected values to be between -1 and 1. Obtained {self.motor_left}, {self.motor_right}"
            raise ValueError(msg)


@dataclass
class Duckiebot1Observations:
    camera: JPGImage


@dataclass
class RGB:
    """ Values between 0, 1. """

    r: float
    g: float
    b: float

    def __post_init__(self):
        for a in [self.r, self.g, self.b]:
            if not isinstance(a, float):
                raise ValueError(a)


@dataclass
class LEDSCommands:
    center: RGB
    front_left: RGB
    front_right: RGB
    back_left: RGB
    back_right: RGB


@dataclass
class Duckiebot1Commands:
    wheels: PWMCommands
    LEDS: LEDSCommands


@dataclass
class DB18SetRobotCommands(SetRobotCommands):
    robot_name: RobotName
    t_effective: float
    commands: Duckiebot1Commands


@dataclass
class DB18RobotObservations(RobotObservations):
    robot_name: RobotName
    t_effective: float
    observations: Duckiebot1Observations


@dataclass
class DTSimRobotInfo:
    pose: np.ndarray
    velocity: np.ndarray
    # last_action: np.ndarray[
    # wheels_velocities: np.ndarray
    pwm: PWMCommands
    leds: LEDSCommands


@dataclass
class DTSimRobotState(RobotState):
    robot_name: RobotName
    t_effective: float
    state: DTSimRobotInfo


@dataclass
class DTSetMap:
    map_data: str


@dataclass
class DTSimState:
    t_effective: float
    duckiebots: Dict[str, DTSimRobotInfo]


@dataclass
class DTSimStateDump(StateDump):
    state: DTSimState


@dataclass
class Duckiebot1ObservationsPlusState:
    camera: JPGImage
    your_name: RobotName
    state: DTSimState
    map_data: str


description = """Particularization for Duckiebot1 observations and commands."""
protocol_agent_duckiebot1 = particularize_no_check(
    protocol_agent,
    description=description,
    inputs={"observations": Duckiebot1Observations},
    outputs={"commands": Duckiebot1Commands},
)

description = """Particularization for Duckiebot1; observations and commands with full state """
protocol_agent_duckiebot1_fullstate = particularize_no_check(
    protocol_agent_duckiebot1, inputs={"observations": Duckiebot1ObservationsPlusState},
)

protocol_simulator_duckiebot1 = particularize_no_check(
    protocol_simulator,
    description="""Particularization for Duckiebot1 observations and commands.""",
    inputs={"set_robot_commands": DB18SetRobotCommands, "set_map": DTSetMap,},
    outputs={
        "robot_observations": DB18RobotObservations,
        "robot_state": DTSimRobotState,
        "state_dump": DTSimStateDump,
    },
)


### DB20


@dataclass
class DB20Odometry:
    resolution_rad: float
    """ What is the resolution of 1 odometry tick"""
    axis_left_rad: float
    """" The current rotation of the left wheel. Positive when robot goes forward """
    axis_right_rad: float
    """" The current rotation of the right wheel. Positive when robot goes forward """


@dataclass
class DB20Observations:
    camera: JPGImage
    odometry: DB20Odometry


@dataclass
class DB20ObservationsPlusState:
    camera: JPGImage
    odometry: DB20Odometry

    your_name: RobotName
    state: DTSimState
    map_data: str


@dataclass
class DB20Commands:
    wheels: PWMCommands
    LEDS: LEDSCommands


@dataclass
class DB20SetRobotCommands(SetRobotCommands):
    robot_name: RobotName
    t_effective: float
    commands: DB20Commands


@dataclass
class DB20RobotObservations(RobotObservations):
    robot_name: RobotName
    t_effective: float
    observations: DB20Observations


protocol_agent_DB20 = particularize_no_check(
    protocol_agent,
    description="""Particularization for DB20 observations and commands.""",
    inputs={"observations": DB20Observations},
    outputs={"commands": DB20Commands},
)

protocol_agent_DB20_fullstate = particularize_no_check(
    protocol_agent_duckiebot1, inputs={"observations": DB20ObservationsPlusState},
)

protocol_simulator_DB20 = particularize_no_check(
    protocol_simulator,
    description="""Particularization for Duckiebot1 observations and commands.""",
    inputs={"set_robot_commands": DB20SetRobotCommands, "set_map": DTSetMap,},
    outputs={
        "robot_observations": DB20RobotObservations,
        "robot_state": DTSimRobotState,
        "state_dump": DTSimStateDump,
    },
)
